from pymongo import DESCENDING, GEOSPHERE, IndexModel

VARIANT_INDEXES = [
    IndexModel([("variant", DESCENDING)]),
    IndexModel([("var_type", DESCENDING)]),
    IndexModel([("not_in_var", DESCENDING)]),
]

PANDAPOWER_INDEXES = {
    "net_bus": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("geo", GEOSPHERE)]),
        IndexModel([("substation", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_line": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_bus", DESCENDING)]),
        IndexModel([("to_bus", DESCENDING)]),
        IndexModel([("geo", GEOSPHERE)]),
        *VARIANT_INDEXES,
    ],
    "net_trafo": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("hv_bus", DESCENDING)]),
        IndexModel([("lv_bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_trafo3w": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("hv_bus", DESCENDING)]),
        IndexModel([("mv_bus", DESCENDING)]),
        IndexModel([("lv_bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_switch": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        IndexModel([("element", DESCENDING)]),
        IndexModel([("et", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_load": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_asymetric_load": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_sgen": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_asymetric_sgen": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_gen": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_ext_grid": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        IndexModel([("junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_shunt": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
    ],
    "net_xward": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_ward": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_motor": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_storage": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("bus", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
}

PANDAPIPES_INDEXES = {
    "net_junction": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("geo", GEOSPHERE)]),
        *VARIANT_INDEXES,
    ],
    "net_pipe": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        IndexModel([("geo", GEOSPHERE)]),
        *VARIANT_INDEXES,
    ],
    "net_valve": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_sink": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_source": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_water_tank": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_flow_control": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_press_control": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_compressor": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_pump": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_circ_pump_mass": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("return_junction", DESCENDING)]),
        IndexModel([("flow_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_circ_pump_pressure": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("return_junction", DESCENDING)]),
        IndexModel([("flow_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_heat_exchanger": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
    "net_heat_consumer": [
        IndexModel([("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)], unique=True),
        IndexModel([("from_junction", DESCENDING)]),
        IndexModel([("to_junction", DESCENDING)]),
        *VARIANT_INDEXES,
    ],
}

COMMON_INDEXES = {
    "net_area": [
        IndexModel(
            [("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)],
            unique=True,
        ),
        IndexModel([("index", DESCENDING)]),
        IndexModel([("buses", DESCENDING)]),
        IndexModel([("lines", DESCENDING)]),
        IndexModel([("connection_points", DESCENDING)]),
        IndexModel([("feeders", DESCENDING)]),
        IndexModel([("trafos", DESCENDING)]),
        IndexModel([("trafo3ws", DESCENDING)]),
        IndexModel([("has_ext_grid", DESCENDING)]),
        IndexModel([("substations", DESCENDING)]),
        IndexModel([("type", DESCENDING)]),
        IndexModel([("substation_buses", DESCENDING)]),
        IndexModel([("level", DESCENDING)]),
        IndexModel([("geo", GEOSPHERE)]),
        *VARIANT_INDEXES,
    ],
    "net_substation": [
        IndexModel(
            [("net_id", DESCENDING), ("index", DESCENDING), ("variant", DESCENDING)],
            unique=True,
        ),
        IndexModel([("index", DESCENDING)]),
        IndexModel([("type", DESCENDING)]),
        IndexModel([("level", DESCENDING)]),
        IndexModel([("geo", GEOSPHERE)]),
        *VARIANT_INDEXES,
    ],
}
MONGODB_INDEXES = PANDAPOWER_INDEXES | PANDAPIPES_INDEXES | COMMON_INDEXES



from contextlib import asynccontextmanager

import uvicorn
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from pandahub.lib.PandaHub import PandaHubError
from pandahub.api.routers import net, projects, timeseries, users, auth, variants
from pandahub.api.internal.db import User, db, AccessToken
from . import pandahub_app_settings as ph_settings
from beanie import init_beanie

@asynccontextmanager
async def lifespan(app: FastAPI):
    await init_beanie(
        database=db,
        document_models=[
            User,
            AccessToken
        ],
    )
    yield

app = FastAPI(lifespan=lifespan)

origins = [
    "http://localhost:8080",
]

if ph_settings.debug:
    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

app.include_router(net.router)
app.include_router(projects.router)
app.include_router(timeseries.router)
app.include_router(users.router)
app.include_router(auth.router)
app.include_router(variants.router)


@app.exception_handler(PandaHubError)
async def pandahub_exception_handler(request: Request, exc: PandaHubError):
    return JSONResponse(
        status_code=exc.status_code,
        content=str(exc),
    )

@app.get("/")
async def ready():
    if ph_settings.debug:
        import os
        return os.environ
    return "Hello World!"


if __name__ == "__main__":
    uvicorn.run("pandahub.api.main:app",
                host=ph_settings.pandahub_server_url,
                port=ph_settings.pandahub_server_port,
                log_level="info",
                reload=True,
                workers=ph_settings.workers)



from typing import Optional, Any

from fastapi import APIRouter, Depends
from pydantic import BaseModel

from pandahub.api.dependencies import pandahub

router = APIRouter(
    prefix="/projects",
    tags=["projects"]
)


# -------------------------
# Projects
# -------------------------

class CreateProject(BaseModel):
    name: str
    settings: Optional[dict] = None

@router.post("/create_project")
def create_project(data: CreateProject, ph=Depends(pandahub)):
    ph.create_project(**data.model_dump(), realm=ph.user_id)
    return {"message": f"Project {data.name} created !"}

class DeleteProject(BaseModel):
    project_id: str
    i_know_this_action_is_final: bool

@router.post("/delete_project")
def delete_project(data: DeleteProject, ph=Depends(pandahub)):
    ph.delete_project(**data.model_dump())
    return True

@router.post("/get_projects")
def get_projects(ph=Depends(pandahub)):
    return ph.get_projects()


class Project(BaseModel):
    name: str

@router.post("/project_exists")
def project_exists(data: Project, ph=Depends(pandahub)):
    return ph.project_exists(**data.model_dump(), realm=ph.user_id)


class SetActiveProjectModel(BaseModel):
    project_name: str

@router.post("/set_active_project")
def set_active_project(data: SetActiveProjectModel, ph=Depends(pandahub)):
    ph.set_active_project(**data.model_dump())
    return str(ph.active_project["_id"])


# -------------------------
# Settings
# -------------------------

class GetProjectSettingsModel(BaseModel):
    project_id: str

@router.post("/get_project_settings")
def get_project_settings(data: GetProjectSettingsModel, ph=Depends(pandahub)):
    settings = ph.get_project_settings(**data.model_dump())
    return settings

class SetProjectSettingsModel(BaseModel):
    project_id: str
    settings: dict

@router.post("/set_project_settings")
def set_project_settings(data: SetProjectSettingsModel, ph=Depends(pandahub)):
    ph.set_project_settings(**data.model_dump())

class SetProjectSettingsValueModel(BaseModel):
    project_id: str
    parameter: str
    value: Any = None

@router.post("/set_project_settings_value")
def set_project_settings_value(data: SetProjectSettingsValueModel, ph=Depends(pandahub)):
    ph.set_project_settings_value(**data.model_dump())

# -------------------------
# Metadata
# -------------------------

class GetProjectMetadataModel(BaseModel):
    project_id: str

@router.post("/get_project_metadata")
def get_project_metadata(data: GetProjectMetadataModel, ph=Depends(pandahub)):
    metadata = ph.get_project_metadata(**data.model_dump())
    return metadata

class SetProjectMetadataModel(BaseModel):
    project_id: str
    metadata: dict

@router.post("/set_project_metadata")
def set_project_metadata(data: SetProjectMetadataModel, ph=Depends(pandahub)):
    return ph.set_project_metadata(**data.model_dump())



import json

from fastapi import APIRouter, Depends
from pydantic import BaseModel

from pandahub.api.dependencies import pandahub

router = APIRouter(
    prefix="/variants",
    tags=["variants"]
)


# -------------------------------
#  ROUTES
# -------------------------------

class GetVariantsModel(BaseModel):
    project_id: str
    net_id: int | str

@router.post("/get_variants")
def get_variants(data: GetVariantsModel, ph=Depends(pandahub)):
    ph.set_active_project_by_id(data.project_id)
    variants_collection = ph.get_project_collection("variant")

    variants = variants_collection.find({"net_id": data.net_id}, projection={"_id": 0})
    return {variant.pop("index"): variant for variant in variants}


class CreateVariantModel(BaseModel):
    project_id: str
    net_id: int
    name: str | None = None
    default_name: str | None = None


class CreateVariantResponseModel(BaseModel):
    net_id: int
    index: int
    name: str | None = None
    date_created: int
    date_changed: int


@router.post("/create_variant")
def create_variant(data: CreateVariantModel, ph=Depends(pandahub)) -> CreateVariantResponseModel:
    ph.set_active_project_by_id(data.project_id)
    return ph.create_variant(net_id=data.net_id, name=data.name, default_name=data.default_name)

class DeleteVariantModel(BaseModel):
    project_id: str
    net_id: int | str
    index: int

@router.post("/delete_variant")
def delete_variant(data: DeleteVariantModel, ph=Depends(pandahub)):
    project_id = data.project_id
    ph.set_active_project_by_id(project_id)
    return ph.delete_variant(data.net_id, data.index)

class UpdateVariantModel(BaseModel):
    project_id: str
    net_id: int | str
    index: int
    data: dict

@router.post("/update_variant")
def update_variant(data: UpdateVariantModel, ph=Depends(pandahub)):
    project_id = data.project_id
    ph.set_active_project_by_id(project_id)
    return ph.update_variant(data.net_id, data.index, data.data)



"""Tests for `spinai_backend` package."""
import copy
import pandas as pd
import numpy as np
import pytest
import datetime
import pandapower.networks as nw
import simbench as sb

code = "1-HV-urban--0--sw"
project = "pytest"

def test_from_tutorial(ph):
    ph.set_active_project(project)
    net = nw.simple_mv_open_ring_net()
    p_mw_profiles = np.random.randint(low=0, high=100, size=(35041, len(net.load))) / 100 * net.load.p_mw.values
    q_mvar_profiles = np.ones((35041, len(net.load)))
    timestamps = pd.date_range(start="01/01/2020", end="31/12/2020", freq="15min")
    p_mw_profiles = pd.DataFrame(p_mw_profiles, index=timestamps)
    weekindex = p_mw_profiles.index[0:(7 * 96)]
    q_mvar_profiles = pd.DataFrame(q_mvar_profiles, index=timestamps)

    # writing the p_mw profile
    ph.write_timeseries_to_db(timeseries=p_mw_profiles[0], netname="simple_mv_open_ring_net",
                           element_index=0, element_type="load",
                           data_type="p_mw", test_kwarg="test_p_mw", collection_name="test_collection")
    # writing the q_mvar profile
    ph.write_timeseries_to_db(timeseries=q_mvar_profiles[0], netname="simple_mv_open_ring_net",
                           element_index=0, element_type="load",
                           data_type="q_mvar", test_kwarg="test_q_mvar", collection_name= "test_collection")

    result = ph.bulk_get_timeseries_from_db({"netname": 'simple_mv_open_ring_net',
                                                "element_type" : "load",
                                                "element_index": 0,
                                                  },
                                                 collection_name="test_collection",
                                                 pivot_by_column="data_type"
                                                 )

    week = ph.get_timeseries_from_db(netname= 'simple_mv_open_ring_net', element_index=0,
                                          element_type="load",
                                          data_type="p_mw",
                                          collection_name='test_collection',
                                          timestamp_range=(datetime.datetime(2020, 1, 1, 0, 0),
                                                           datetime.datetime(2020, 1, 8, 0, 0)))

    assert (result.keys() == ['p_mw', 'q_mvar']).all()
    assert result.size == 70082
    assert result.index.dtype == "<M8[ns]"
    assert np.isclose(result.p_mw.sum(), p_mw_profiles.sum()[0])
    assert len(week) == 672
    # 17348.07

def test_simbench_sinlge_ts(ph):
    ph.set_active_project(project)
    net2 = sb.get_simbench_net(code)
    profiles = net2.profiles["load"].set_index('time')
    i = net2.load.index[0]
    element_type = 'load'
    data_type = 'p_mw'
    collection = 'test_collection'
    p_mw_profile = profiles[net2.load.profile[i] + "_pload"]
    ph.write_timeseries_to_db(p_mw_profile, netname=code,
                                    element_index=int(i),
                                    element_type=element_type,
                                    data_type=data_type,
                                    collection_name=collection)

    read_profile = ph.get_timeseries_from_db(netname=code,
                                                   element_index=int(i),
                                                   element_type=element_type,
                                                   data_type=data_type,
                                                   collection_name=collection)

    assert p_mw_profile.sum() == read_profile.sum()
    meta = ph.get_timeseries_metadata({"netname": code,
                                             "element_type": element_type,
                                    },
                                   collection_name=collection)

    assert meta.iloc[0].element_type == element_type
    assert meta.iloc[0].data_type == data_type
    assert meta.iloc[0].netname == code
    assert meta.iloc[0].first_timestamp == p_mw_profile.index[0]
    assert meta.iloc[0].last_timestamp == p_mw_profile.index[-1]

def test_simbench_loop_ts(ph):
    ph.set_active_project(project)
    net = sb.get_simbench_net(code)
    profiles = net.profiles["load"].set_index('time')

    for i in net.load.index[:3]:
        i = net.load.index[0]
        p_mw_profile = profiles[net.load.profile[i] + "_pload"]#.astype(float)

        ph.write_timeseries_to_db(p_mw_profile, netname=code, element_index=int(i), element_type='load',
                               data_type='p_mw', collection_name='test_collection')

        read_profile = ph.get_timeseries_from_db(netname=code, element_index=int(i), element_type='load',
                                       data_type='p_mw', collection_name='test_collection')

        assert p_mw_profile.sum() == read_profile.sum()


def test_del_single_ts_on_db(ph):
    ph.set_active_project(project)
    # write ts to db
    net2 = sb.get_simbench_net(code)
    profiles = net2.profiles["load"].set_index('time')
    i = net2.load.index[0]
    p_mw_profile = profiles[net2.load.profile[i] + "_pload"]#.astype(float)
    ph.write_timeseries_to_db(p_mw_profile, netname=code, element_index=int(i), element_type='load',
                               data_type='p_mw', collection_name='test_collection')

    # read ts from db
    read_profile = ph.get_timeseries_from_db(netname=code, element_index=int(i), element_type='load',
                                   data_type='p_mw', collection_name='test_collection')

    assert p_mw_profile.sum() == read_profile.sum()

    # delete ts

    ph.delete_timeseries_from_db(element_type='load',
                                   data_type='p_mw',
                                   element_index=int(i),
                                   collection_name="test_collection",
                                   netname=code)

    # read again
    try:
        ph.get_timeseries_from_db(netname=code,
                                            element_index=int(i),
                                            element_type='load',
                                            data_type='p_mw',
                                            collection_name='test_collection')
        assert False # this line shouldnt be reached, because the function triggers KeyError when no timeseries is found
    except:
        assert True


def test_bulk_ts_on_db(ph):
    ph.set_active_project(project)
    # write bulk cts to db
    p_mw_profiles = np.random.randint(low=0, high=100, size=(35041, 10)).astype(float)
    timestamps = pd.date_range(start="01/01/2020", end="31/12/2020", freq="15min")
    p_mw_profiles_no_time = pd.DataFrame(copy.deepcopy(p_mw_profiles))
    p_mw_profiles = pd.DataFrame(p_mw_profiles, index=timestamps)

    ph.bulk_write_timeseries_to_db(p_mw_profiles,
                                   element_type="load",
                                   data_type="p_mw",
                                   netname="bulk_write_net",
                                   collection_name="test_collection")

    result = ph.bulk_get_timeseries_from_db({"netname": 'bulk_write_net', "element_type": "load",},
                                            collection_name="test_collection",
                                            pivot_by_column="element_index")



    bulk_week = ph.bulk_get_timeseries_from_db({"netname": 'bulk_write_net', "element_type": "load",},
                                               collection_name="test_collection",
                                               pivot_by_column="element_index",
                                               timestamp_range=(datetime.datetime(2020, 1, 1, 0, 0),
                                                                datetime.datetime(2020, 1, 8, 0, 0)))


    p_mw_profiles_no_time = pd.DataFrame(p_mw_profiles_no_time)
    ph.bulk_write_timeseries_to_db(p_mw_profiles_no_time, element_type="sgen", data_type="p_mw",
                                        netname="bulk_write_net", collection_name="test_collection")

    result_no_time = ph.bulk_get_timeseries_from_db({"netname": 'bulk_write_net',
                                                  "element_type": "sgen",
                                                  }, collection_name="test_collection",
                                                 pivot_by_column="element_index"
                                                 )

    assert bulk_week.size == 6720
    assert result_no_time.size == 350410
    assert result.sum().sum() == p_mw_profiles.sum().sum()


    ph.bulk_del_timeseries_from_db({"netname": 'bulk_write_net',
                                                  "element_type": "load",
                                                  }, collection_name="test_collection")

    del_res = ph.bulk_get_timeseries_from_db({"netname": 'bulk_write_net',
                                                  "element_type": "load",
                                                  }, collection_name="test_collection"
                                                 , pivot_by_column="element_index"

                                                 )

    assert len(del_res) == 0

def test_add_metadata(ph):
    ph.set_active_project(project)
    p_mw_profiles = np.random.randint(low=0, high=100, size=(25, 10)).astype(float)
    timestamps = pd.date_range(start="01/01/2020", end="01/02/2020",
                               freq="60min")
    p_mw_profiles = pd.DataFrame(p_mw_profiles, index=timestamps)
    ph.write_timeseries_to_db(p_mw_profiles[0],
                                    netname="test_add_metadata",
                                    element_index=0,
                                    element_type="load",
                                    data_type="p_mw",
                                    test_kwarg="test_p_mw",
                                    collection_name= "test_collection")

    result = ph.get_timeseries_from_db(netname= 'test_add_metadata',
                                             element_index=0,
                                             element_type="load",
                                             data_type="p_mw",
                                             collection_name='test_collection')
    filter = {"netname": 'test_add_metadata',
                "element_type" : "load",
                "element_index": 0,
                 }

    meta_before = ph.get_timeseries_metadata(filter, collection_name="test_collection")

    add_meta = {"max": str(result.values.max())}

    ph.add_metadata(filter, add_meta=add_meta,
                          collection_name="test_collection")
    # check for new metadata
    meta_after = ph.get_timeseries_metadata(filter, collection_name="test_collection")

    assert len(meta_after.columns) == 13
    assert len(meta_before.columns) == 12

def test_bulk_write_with_meta(ph):
    ph.set_active_project(project)
    p_mw_profiles = np.random.randint(low=0, high=100, size=(97, 10)).astype(float)
    timestamps = pd.date_range(start="01/01/2020", end="01/02/2020", freq="15min")
    p_mw_profiles = pd.DataFrame(p_mw_profiles, index=timestamps)
    meta = pd.DataFrame(p_mw_profiles.max(), columns=["max"], dtype=object)

    ph.bulk_write_timeseries_to_db(p_mw_profiles, element_type="meta_test_load", data_type="p_mw",
                                netname="bulk_write_net", collection_name="test_collection",
                                meta_frame=meta)

    result = ph.bulk_get_timeseries_from_db({"netname": 'bulk_write_net',
                                                     "element_type": "meta_test_load",
                                                     },
                                         collection_name="test_collection",
                                         pivot_by_column="element_index")

    bulk_meta = ph.get_timeseries_metadata({"netname": 'bulk_write_net',
                                                  "element_type": "meta_test_load",
                                                  },
                                                 collection_name="test_collection")

    assert len(bulk_meta) == 10



if __name__ == '__main__':
    from pandahub import PandaHub

    ph = PandaHub()

    project_name = "pytest"

    if ph.project_exists(project_name):
        ph.set_active_project(project_name)
        ph.delete_project(i_know_this_action_is_final=True)

    ph.create_project(project_name)
    ph.set_active_project(project_name)

    p_mw_profiles = np.random.randint(low=0, high=100, size=(24, 10))
    timestamps = pd.date_range(start="01/01/2020", end="01/02/2020",
                               freq="60min", closed="right")
    p_mw_profiles = pd.DataFrame(p_mw_profiles, index=timestamps)
    ph.write_timeseries_to_db(p_mw_profiles[0],
                                    netname="test_add_metadata",
                                    element_index=0,
                                    element_type="load",
                                    data_type="p_mw",
                                    test_kwarg="test_p_mw",
                                    collection_name= "test_collection")

    result = ph.get_timeseries_from_db(netname= 'test_add_metadata',
                                             element_index=0,
                                             element_type="load",
                                             data_type="p_mw",
                                             collection_name='test_collection')
    filter = {"netname": 'test_add_metadata',
                "element_type" : "load",
                "element_index": 0,
                 }

    meta_before = ph.get_timeseries_metadata(filter, collection_name="test_collection")

    add_meta = {"max": str(result.values.max())}

    ph.add_metadata(filter, add_meta=add_meta, collection_name="test_collection")
    # check for new metadata
    meta_after = ph.get_timeseries_metadata(filter, collection_name="test_collection")

    assert len(meta_after.columns) == 12
    assert len(meta_before.columns) == 11




import numpy as np
import datetime
from pandahub.mongo_io_methods import MongoIOMethods
from pandapower.timeseries.data_source import DataSource
import pandas as pd

try:
    import pplog
    logger = pplog.getLogger(__name__)
except ImportError:
    import logging


class MongoData(DataSource):
    """
    Fetches timeseries data from a mongodb
    """

    def __init__(self, io_methods: MongoIOMethods, netname: str, db_name: str, element_index: list,
                 data_type='p_mw', element_type='load', collection_name='timeseries_data', prefetch_count=1000,
                 **kwargs):

        super(MongoData, self).__init__()
        self.db_name = db_name
        self.collection_name = collection_name

        if type(element_index) is pd.Int64Index:
            element_index = element_index.values.tolist()

        filter = {
            "netname": netname,
            "data_type": data_type,
            "element_type": element_type,
            # "element_index": element_index
        }
        self.filter = {**filter, **kwargs}

        self.io_methods = io_methods

        self.tseries = None
        self.prefetch_count = prefetch_count
        self.current_fetch_position = -1
        self.metadata = self.io_methods.get_timeseries_metadata(filter_document=self.filter, db_name=self.db_name,
                                                                collection_name=self.collection_name)

        first_timestamp = self.metadata['first_timestamp'].values[0]

        if type(first_timestamp) == str:
            self.first_timestamp = datetime.datetime.fromisoformat(first_timestamp)
        elif type(first_timestamp) == np.datetime64:
            self.first_timestamp = pd.Timestamp(first_timestamp).to_pydatetime()
        else:
            self.first_timestamp = first_timestamp

    def get_time_step_value(self, time_step, profile_name, scale_factor=1.0):
        fs = self.first_timestamp + datetime.timedelta(minutes=15 * time_step)
        if time_step >= self.current_fetch_position:
            self.current_fetch_position = time_step + self.prefetch_count
            # fs = self.first_timestamp + datetime.timedelta(minutes=15*time_step)
            es = self.first_timestamp + datetime.timedelta(minutes=15*self.current_fetch_position)
            self.tseries = self.io_methods.bulk_get_timeseries_from_db(filter_document=self.filter, db_name=self.db_name,
                                                                       collection_name=self.collection_name,
                                                                       timestamp_range=[fs, es],
                                                                       pivot_by_column="element_index")

        try:
            return self.tseries.loc[fs.isoformat(), profile_name] * scale_factor
        except KeyError:
            if type(profile_name) == pd.Int64Index:
                self.tseries.columns = self.tseries.columns.astype(int)
                return self.tseries.loc[fs.isoformat(), profile_name] * scale_factor
            elif type(profile_name) == list:
                self.tseries.columns = self.tseries.columns.astype(str)
                t = self.tseries.loc[fs.isoformat(), profile_name] * scale_factor
                t.index = t.index.astype(int)
                return t



from fastapi import APIRouter

from pandahub.api.internal.schemas import UserRead, UserUpdate
from pandahub.api.internal.users import fastapi_users

router = APIRouter(
    prefix="/users",
    tags=["users"]
)

router.include_router(fastapi_users.get_users_router(UserRead, UserUpdate))



import json

import pandas as pd
from fastapi import APIRouter, Depends
from pydantic import BaseModel
from typing import Optional

from pandahub.api.dependencies import pandahub

router = APIRouter(prefix="/timeseries", tags=["timeseries"])


# -------------------------------
#  ROUTES
# -------------------------------


class GetTimeSeriesModel(BaseModel):
    filter_document: Optional[dict] = {}
    global_database: Optional[bool] = False
    project_id: Optional[str] = None
    timestamp_range: Optional[tuple] = None
    exclude_timestamp_range: Optional[tuple] = None
    collection_name: Optional[str] = "timeseries"


@router.post("/get_timeseries_from_db")
def get_timeseries_from_db(data: GetTimeSeriesModel, ph=Depends(pandahub)):
    if data.timestamp_range is not None:
        data.timestamp_range = [pd.Timestamp(t) for t in data.timestamp_range]
    ts = ph.get_timeseries_from_db(**data.model_dump())
    return ts.to_json(date_format="iso")


class MultiGetTimeSeriesModel(BaseModel):
    filter_document: Optional[dict] = {}
    global_database: Optional[bool] = False
    project_id: Optional[str] = None
    timestamp_range: Optional[tuple] = None
    exclude_timestamp_range: Optional[tuple] = None
    collection_name: Optional[str] = "timeseries"


@router.post("/multi_get_timeseries_from_db")
def multi_get_timeseries_from_db(data: MultiGetTimeSeriesModel, ph=Depends(pandahub)):
    if data.timestamp_range is not None:
        data.timestamp_range = [pd.Timestamp(t) for t in data.timestamp_range]
    ts = ph.multi_get_timeseries_from_db(**data.model_dump(), include_metadata=True)
    for i, data in enumerate(ts):
        ts[i]["timeseries_data"] = data["timeseries_data"].to_json(date_format="iso")
    return ts


class GetTimeseriesMetadataModel(BaseModel):
    project_id: str
    filter_document: Optional[dict] = {}
    global_database: Optional[bool] = False
    collection_name: Optional[str] = "timeseries"


@router.post("/get_timeseries_metadata")
def get_timeseries_metadata(data: GetTimeseriesMetadataModel, ph=Depends(pandahub)):
    ph.set_active_project_by_id(data.project_id)
    ts = ph.get_timeseries_metadata(
        filter_document=data.filter_document,
        global_database=data.global_database,
        collection_name=data.collection_name,
    )
    ts = json.loads(ts.to_json(orient="index"))
    return ts


class WriteTimeSeriesModel(BaseModel):
    timeseries: str
    project_id: Optional[str] = None
    data_type: Optional[str] = None
    element_type: Optional[str] = None
    netname: Optional[str] = None
    element_index: Optional[int] = None
    global_database: Optional[bool] = False
    collection_name: Optional[str] = "timeseries"
    name: Optional[str] = None


@router.post("/write_timeseries_to_db")
def write_timeseries_to_db(data: WriteTimeSeriesModel, ph=Depends(pandahub)):
    data.timeseries = pd.Series(json.loads(data.timeseries))
    data.timeseries.index = pd.to_datetime(data.timeseries.index)
    ph.write_timeseries_to_db(**data.model_dump())
    return True



import uuid

from fastapi_users import schemas
from .. import pandahub_app_settings as ph_settings


class UserRead(schemas.BaseUser[uuid.UUID]):
    pass


class UserCreate(schemas.BaseUserCreate):
    is_active: bool = not ph_settings.registration_admin_approval


class UserUpdate(schemas.BaseUserUpdate):
    pass



import uuid
from fastapi import Depends, Request
from fastapi_users import BaseUserManager, FastAPIUsers, UUIDIDMixin
from fastapi_users.authentication import AuthenticationBackend, BearerTransport
from fastapi_users.authentication.strategy.db import (
    AccessTokenDatabase,
    DatabaseStrategy,
)
from fastapi_users.db import BeanieUserDatabase

from ..internal.db import get_user_db, get_access_token_db, User, AccessToken

class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
  pass

async def get_user_manager(user_db:BeanieUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)

bearer_transport = BearerTransport(tokenUrl="auth/login")


def get_database_strategy(
    access_token_db: AccessTokenDatabase[AccessToken] = Depends(get_access_token_db),
) -> DatabaseStrategy:
    return DatabaseStrategy(access_token_db)


auth_backend = AuthenticationBackend(
    name="database",
    transport=bearer_transport,
    get_strategy=get_database_strategy,
)
fastapi_users = FastAPIUsers(get_user_manager, [auth_backend])



import os
from typing import Annotated

from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic import AfterValidator


def get_secret(secret: str) -> str:
    """If arg is a file path, read and return file content."""
    if secret and os.path.isfile(secret):
        with open(secret) as f:
            secret = f.read()
    return secret


SecretFromFile = Annotated[str, AfterValidator(get_secret)]


class PandaHubSettings(BaseSettings):
    """PandaHub settings"""

    model_config = SettingsConfigDict(env_ignore_empty=True)

    mongodb_url: str = "mongodb://localhost:27017"
    mongodb_user: SecretFromFile | None = None
    mongodb_password: SecretFromFile | None = None
    mongodb_global_database_url: str | None = None
    mongodb_global_database_user: str | None = None
    mongodb_global_database_password: str | None = None
    pandahub_global_db_client: bool = False
    create_indexes_with_project: bool = True


pandahub_settings = PandaHubSettings()



import asyncio
import uuid

import motor.motor_asyncio
from beanie import Document
from fastapi_users.db import BeanieBaseUser, BeanieUserDatabase
from fastapi_users_db_beanie.access_token import (
    BeanieAccessTokenDatabase,
    BeanieBaseAccessToken,
)

from .. import pandahub_app_settings as ph_settings
from pydantic import Field

mongo_client_args = {"host": ph_settings.mongodb_url, "uuidRepresentation": "standard", "connect": False}
if ph_settings.mongodb_user:
    mongo_client_args |= {"username": ph_settings.mongodb_user, "password": ph_settings.mongodb_password}

client = motor.motor_asyncio.AsyncIOMotorClient(**mongo_client_args)
client.get_io_loop = asyncio.get_event_loop
db = client["user_management"]

class User(BeanieBaseUser, Document):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    is_active: bool = not ph_settings.registration_admin_approval
    class Settings(BeanieBaseUser.Settings):
        name = "users"


class AccessToken(BeanieBaseAccessToken, Document):
    user_id: uuid.UUID = Field(default_factory=uuid.uuid4)
    class Settings(BeanieBaseAccessToken.Settings):
        name = "access_tokens"

async def get_user_db():
    yield BeanieUserDatabase(User)

async def get_access_token_db():
    yield BeanieAccessTokenDatabase(AccessToken)



import pytest

import pandapipes as pps
import pandapipes.networks as nw_pps
import pandapower as pp
import pandapower.networks as nw_pp
from pandahub import PandaHubError
from pandapipes.toolbox import nets_equal


def test_additional_res_tables(ph):
    import pandas as pd
    ph.set_active_project("pytest")

    # reset project aka delete everything
    db = ph._get_project_database()
    for cname in db.list_collection_names():
        db.drop_collection(cname)

    net1 = pp.create_empty_network()
    net1['res_test'] = pd.DataFrame(data={'col1': [1, 2], 'col2': [3, 4]})
    ph.write_network_to_db(net1, 'test')
    net2 = ph.get_network_by_name('test')

    assert('res_test' in net2)
    assert(net1.res_test.shape == (2,2))


def test_network_io(ph):
    ph.set_active_project("pytest")
    # reset project aka delete everything
    db = ph._get_project_database()
    for cname in db.list_collection_names():
        db.drop_collection(cname)

    net1 = nw_pp.mv_oberrhein()
    name1 = "oberrhein_network"
    # for some unknown reason the format of mv_oberrhein does not match the latest pandapower format
    # net1.gen.rename(columns={"qmax_mvar": "max_q_mvar", "qmin_mvar": "min_q_mvar"}, inplace=True)
    # del net1.shunt["scaling"]
    # del net1.impedance["r_pu"]
    # del net1.impedance["x_pu"]
    # del net1.dcline["cost_per_mw"]

    net2 = nw_pp.simple_four_bus_system()
    pp.create_bus(net2, vn_kv=20, index=10)  # check non-consecutive indices
    name2 = "simple_network"

    # we check storing two different networks consecutively to ensure the nets are properly separated
    for net, name in [(net1, name1), (net2, name2)]:
        if ph.network_with_name_exists(name):
            ph.delete_network_by_name(name)

        assert not ph.network_with_name_exists(name)

        ph.write_network_to_db(net, name)
        assert ph.network_with_name_exists(name) == True

        net_loaded = ph.get_network_by_name(name)

        pp.runpp(net)
        pp.runpp(net_loaded)

        # This will fail if the net contains 'None' values. Since they get casted to NaN which by definition
        # doesn't compare to itself
        # assert pp.nets_equal(net, net_loaded, check_dtype=False)

        net3 = ph.get_network_by_name(name, only_tables=["bus"])
        assert len(net3.bus) == len(net.bus)
        assert len(net3.line) == 0
        assert len(net3.load) == 0

    # delete first network
    ph.delete_network_by_name(name1)
    assert ph.network_with_name_exists(name1) == False

    # check that second network is still in database
    assert ph.network_with_name_exists(name2) == True
    net2_loaded = ph.get_network_by_name(name2)
    pp.runpp(net2_loaded)
    assert pp.nets_equal(net2, net2_loaded, check_only_results=True)


def test_load_subnetwork(ph):
    ph.set_active_project("pytest")
    name = "oberrhein_network"

    if not ph.network_with_name_exists(name):
        net = nw_pp.mv_oberrhein()
        ph.write_network_to_db(net, name)

    subnet = ph.get_subnet_by_name(name, node_filter={"vn_kv": 110})
    expected_sizes = [("bus", 4), ("line", 0), ("trafo", 2), ("ext_grid", 2)]

    for element, size in expected_sizes:
        assert len(subnet[element]) == size
        assert len(subnet["res_" + element]) == size

    subnet = ph.get_subnet_by_name(name, node_filter={"vn_kv": 110},
                                   include_results=False)
    for element, size in expected_sizes:
        assert len(subnet[element]) == size
        assert len(subnet["res_" + element]) == 0

    subnet = ph.get_subnet_by_name(name, node_filter={"vn_kv": 110},
                                   add_edge_branches=False)
    expected_sizes = [("bus", 2), ("line", 0), ("trafo", 0), ("ext_grid", 2)]

    for element, size in expected_sizes:
        assert len(subnet[element]) == size
        assert len(subnet["res_" + element]) == size


def test_access_and_set_single_values(ph):
    ph.set_active_project("pytest")
    name = "oberrhein_network"

    net = nw_pp.mv_oberrhein()
    if not ph.network_with_name_exists(name):
        ph.write_network_to_db(net, name)

    element = "sgen"
    parameter = "p_mw"
    index = 4
    p_mw_new = 0.1

    value = ph.get_net_value_from_db(name, element, index, parameter)
    assert value == net[element][parameter].at[index]

    ph.set_net_value_in_db(name, element, index, parameter, p_mw_new)
    value = ph.get_net_value_from_db(name, element, index, parameter)
    assert value == p_mw_new

    ph.delete_element(name, element, index)
    with pytest.raises(PandaHubError):
        ph.get_net_value_from_db(name, element, index, parameter)
    net = ph.get_network_by_name(name)
    assert index not in net[element].index


def test_pandapipes(ph):
    ph.set_active_project('pytest')
    net = nw_pps.gas_versatility()
    ph.write_network_to_db(net, 'versatility')
    net2 = ph.get_network_by_name('versatility', convert=False)
    pps.pipeflow(net)
    pps.pipeflow(net2)
    assert nets_equal(net, net2, check_only_results=True)


def test_get_set_single_value(ph):
    ph.set_active_project('pytest')
    net = nw_pp.mv_oberrhein()
    ph.write_network_to_db(net, 'oberrhein')
    val = ph.get_net_value_from_db('oberrhein', 'load', 0, 'p_mw')
    assert val == net.load.at[0, 'p_mw']
    ph.set_net_value_in_db('oberrhein', 'load', 0, 'p_mw', 0.5)
    val = ph.get_net_value_from_db('oberrhein', 'load', 0, 'p_mw')
    assert val == 0.5


if __name__ == '__main__':
    from pandahub import PandaHub

    ph = PandaHub()
    ph.create_project('pytest')
    net = nw_pps.gas_versatility()
    ph.write_network_to_db(net, 'versatility')
    net2 = ph.get_network_by_name('versatility')
    pps.pipeflow(net)
    pps.pipeflow(net2)
    # test_network_io(ph)
    # 0 / 0
    # project_name = "pytest"

    # if ph.project_exists(project_name):
    #     ph.set_active_project(project_name)
    #     ph.delete_project(i_know_this_action_is_final=True)

    # ph.create_project(project_name)
    # ph.set_active_project("pytest")
    # net = ph.get_net_from_db("oberrhein_network")
    # name = "oberrhein_network"

    # net = nw.mv_oberrhein()
    # if not ph.network_with_name_exists(name):
    #     ph.write_network_to_db(net, name)

    # element = "sgen"
    # parameter = "p_mw"
    # index = 4
    # p_mw_new = 0.1

    # value = ph.get_net_value_from_db(name, element, index, parameter)
    # assert value == net[element][parameter].at[index]

    # ph.set_net_value_in_db(name, element, index, parameter, p_mw_new)
    # value = ph.get_net_value_from_db(name, element, index, parameter)
    # assert value == p_mw_new

    # ph.delete_net_element(name, element, index)
    # with pytest.raises(PandaHubError):
    #     ph.get_net_value_from_db(name, element, index, parameter)
    # net = ph.get_net_from_db(name)
    # assert index not in net[element].index



from contextlib import contextmanager
from typing import Generator, overload

from pymongo import MongoClient
from pymongo.synchronous.collection import Collection
from pymongo.synchronous.database import Database

from pandahub.lib.settings import pandahub_settings as settings


def _get_mongo_client(
    connection_url: str = settings.mongodb_url,
    connection_user: str = settings.mongodb_user,
    connection_password: str = settings.mongodb_password,
    background_connect: bool = True,
) -> MongoClient:
    mongo_client_args = {
        "host": connection_url,
        "uuidRepresentation": "standard",
        "connect": background_connect,
    }
    if connection_user:
        mongo_client_args |= {
            "username": connection_user,
            "password": connection_password,
        }
    return MongoClient(**mongo_client_args)

if settings.pandahub_global_db_client:
    _global_mongo_client = _get_mongo_client(connection_url=settings.mongodb_url, connection_user = settings.mongodb_user,
                                             connection_password = settings.mongodb_password, background_connect=False)
else:
    _global_mongo_client = None




@overload
def get_mongo_client(database: None = None, collection: None = None,
                     connection_url: str = ..., connection_user: str = ..., connection_password: str = ...) -> MongoClient: ...
@overload
def get_mongo_client(database: str, collection: None=None,
                     connection_url: str = ..., connection_user: str = ..., connection_password: str = ...) -> Database: ...
@overload
def get_mongo_client(database: str, collection: str,
                     connection_url: str = ..., connection_user: str = ..., connection_password: str = ...) -> Collection: ...

def get_mongo_client(
    database: str | None = None,
    collection: str | None = None,
    connection_url: str = settings.mongodb_url,
    connection_user: str = settings.mongodb_user,
    connection_password: str = settings.mongodb_password,
) -> MongoClient | Database | Collection:
    if collection is not None and database is None:
        raise ValueError("Must specify database to access a collection!")
    if _global_mongo_client is None:
        client = _get_mongo_client(connection_url, connection_user, connection_password)
    else:
        client = _global_mongo_client
    return _get_db_or_coll(client, database, collection)


@contextmanager
def mongo_client(
    database: str | None = None,
    collection: str | None = None,
    connection_url: str = settings.mongodb_url,
    connection_user: str = settings.mongodb_user,
    connection_password: str = settings.mongodb_password,
) -> Generator[MongoClient | Database | Collection, None, None]:
    """Contextmanager for pymongo MongoClient / Database / Collection with close after use.

    Parameters
    ----------
    database: str or None
        The database to connect to
    collection: str or None
        The collection to connect to
    connection_url: str or None
        Defaults to MONGODB_URL env var
    connection_user: str or None
        Defaults to MONGODB_USER env var
    connection_password: str or None
        Defaults to MONGODB_PASSWORD env var

    Returns
    -------
        Contextmanager yielding MongoClient / Database / Collection
    """
    if collection is not None and database is None:
        raise ValueError("Must specify database to access a collection!")
    client = _get_mongo_client(connection_url, connection_user, connection_password)
    try:
        yield _get_db_or_coll(client, database, collection)
    finally:
        client.close()


def _get_db_or_coll(client: MongoClient, database: str | None = None, collection: str | None = None) -> MongoClient | Database | Collection:
    if database is not None and collection is not None:
        return client[database][collection]
    elif database is not None:
        return client[database]
    else:
        return client



from fastapi import Depends

from pandahub import PandaHub
from .internal.db import User
from .internal.users import fastapi_users

current_active_user = fastapi_users.current_user(active=True)


def pandahub(user: User = Depends(current_active_user)):
    return PandaHub(user_id=str(user.id))



from pydantic_settings import SettingsConfigDict

from pandahub.lib.settings import PandaHubSettings


class PandaHubAppSettings(PandaHubSettings):
    """PandaHub app settings"""

    model_config = SettingsConfigDict(env_ignore_empty=True)

    registration_enabled: bool = True
    registration_admin_approval: bool = False
    pandahub_server_url: str = "0.0.0.0"
    pandahub_server_port: int =8002
    workers: int = 2
    debug: bool = False






import numpy as np
import pandas as pd
from datetime import datetime
from types import FunctionType
from pandapower.timeseries import OutputWriter
from pandahub.mongo_io_methods import MongoIOMethods

try:
    import pplog
    logger = pplog.getLogger(__name__)
except ImportError:
    import logging


class OutputWriterMongoDB(OutputWriter):
    """
    Output Writer which writes to a mongoDB
    """
    def __init__(self, net, io_methods: MongoIOMethods, netname: str, db_name: str, start_date: datetime, time_steps=None,
                 write_time=None, log_variables=None, write_caching=10, freq="15min", collection_name='timeseries_data',
                 **kwargs):
        super().__init__(net, time_steps=time_steps, write_time=write_time, log_variables=log_variables)
        self.io_methods = io_methods
        self.args = kwargs
        self.NET_NAME = netname
        self.db_name = db_name
        self.write_caching = write_caching
        self.current_pos = 0
        self.ids = dict()
        self.collection_name = collection_name
        self.output = dict()
        self.freq = freq
        self.start_date = start_date

    # def dump_to_file(self, net, append=False, recycle_options=None):
    #     pass

    # def _save_single_xls_sheet(self, append):
        # ToDo: implement save to a single sheet
    #     raise NotImplementedError("Sorry not implemented yet")

    def _init_np_array(self, partial_func):
        (table, variable, net, index, eval_function, eval_name) = partial_func.args
        hash_name = self._get_np_name(partial_func.args)
        n_columns = len(index)
        if eval_function is not None:
            n_columns = 1
            if isinstance(eval_function, FunctionType):
                if "n_columns" in eval_function.__code__.co_varnames:
                    n_columns = eval_function.__defaults__[0]
        # self.np_results[hash_name] = np.zeros((len(self.time_steps), n_columns))
        self.np_results[hash_name] = np.zeros((self.write_caching, n_columns))

    def _log(self, table, variable, net, index, eval_function=None, eval_name=None):
        try:
            # ToDo: Create a mask for the numpy array in the beginning and use this one for getting the values. Faster
            if net[table].index.equals(pd.Index(index)):
                # if index equals all values -> get numpy array directly
                result = net[table][variable].values
            else:
                # get by loc (slow)
                result = net[table].loc[index, variable].values

            if eval_function is not None:
                result = eval_function(result)

            # save results to numpy array
            # time_step_idx = self.time_step_lookup[self.time_step]
            hash_name = self._get_np_name((table, variable, net, index, eval_function, eval_name))
            # self.np_results[hash_name][time_step_idx, :] = result
            self.np_results[hash_name][self.current_pos, :] = result

        except Exception as e:
            logger.error("Error at index %s for %s[%s]: %s" % (index, table, variable, e))

    def _np_to_pd(self):
        # convert numpy arrays (faster so save results) into pd Dataframes (user friendly)
        # intended use: At the end of time series simulation write results to pandas
        res_df = dict()

        for partial_func in self.output_list:
            (table, variable, net, index, eval_func, eval_name) = partial_func.args
            # res_name = self._get_hash(table, variable)
            res_name = self._get_output_name(table, variable)
            np_name = self._get_np_name(partial_func.args)
            columns = index
            if eval_name is not None and eval_func is not None:
                if isinstance(eval_func, FunctionType):
                    if "n_columns" not in eval_func.__code__.co_varnames:
                        columns = [eval_name]
                else:
                    columns = [eval_name]

            # res_df = pd.DataFrame(self.np_results[np_name], index=self.time_steps, columns=columns)
            res_df[res_name] = pd.DataFrame(self.np_results[np_name], columns=columns)

        return res_df

    def save_results(self, net, time_step, pf_converged, ctrl_converged, recycle_options=None):
        # remember the last time step
        self.time_step = time_step

        if not pf_converged:
            super().save_nans_to_parameters()
            self.output["Parameters"].loc[time_step, "powerflow_failed"] = True
        elif not ctrl_converged:
            self.output["Parameters"].loc[time_step, "controller_unstable"] = True
        else:
            super().save_to_parameters()

        res = self._np_to_pd()
        write_to_db = False

        self.current_pos += 1
        if self.current_pos >= self.write_caching:
            write_to_db = True

        # last time_step
        if self.time_step == self.time_steps[-1]:
            print(time_step)
            write_to_db = True

        if write_to_db:
            for res_name, res_df in res.items():
                end = self.start_date + (self.current_pos - 1) * pd.Timedelta(self.freq)
                if self.current_pos < self.write_caching:
                    res_df.drop(range(self.current_pos, self.write_caching), axis=0, inplace=True)
                res_df.index = pd.date_range(start=self.start_date,
                                             end=end,
                                             freq=self.freq)
                if res_name in self.ids:
                    for ii in res_df.index:
                        row = res_df.loc[ii]
                        self.io_methods.bulk_update_timeseries_in_db(
                            new_ts_content=pd.DataFrame(row).transpose(),
                            document_ids=self.ids[res_name],
                            db_name=self.db_name,
                            collection_name=self.collection_name,
                            # **self.args
                        )
                else:
                    et = res_name.split('.')[0]
                    dt = res_name.split('.')[1]
                    self.ids[res_name] = self.io_methods.bulk_write_timeseries_to_db(
                        timeseries=res_df, netname=self.NET_NAME,
                        element_type=et, data_type=dt,
                        collection_name=self.collection_name,
                        db_name=self.db_name,
                        return_ids=True,
                        last_timestamp=self.start_date + pd.Timedelta(self.freq) * len(self.time_steps),
                        num_timestamps=len(self.time_steps),
                        **self.args
                    )

            self.start_date = self.start_date + self.current_pos * pd.Timedelta(self.freq)
            self.current_pos = 0



from typing import Optional, Any

import pandapower as pp
import pandapipes as pps
from fastapi import APIRouter, Depends
from pydantic import BaseModel

from pandahub.api.dependencies import pandahub

router = APIRouter(
    prefix="/net",
    tags=["net"]
)


# -------------------------
# Net handling
# -------------------------

class GetNetFromDB(BaseModel):
    project_id: str
    name: str
    include_results: bool
    only_tables: Optional[list] = None


@router.post("/get_net_from_db")
def get_net_from_db(data: GetNetFromDB, ph=Depends(pandahub)):
    net = ph.get_network_by_name(**data.model_dump())
    return pp.to_json(net)


class WriteNetwork(BaseModel):
    project_id: str
    net: str
    name: str
    overwrite: Optional[bool] = True


@router.post("/write_network_to_db")
def write_network_to_db(data: WriteNetwork, ph=Depends(pandahub)):
    params = data.model_dump()
    params["net"] = pp.from_json_string(params["net"])
    ph.write_network_to_db(**params)


# -------------------------
# Element CRUD
# -------------------------

class BaseCRUDModel(BaseModel):
    project_id: str
    net_id: int | str
    element_type: str

class GetNetValueModel(BaseCRUDModel):
    element_index: int
    parameter: str

@router.post("/get_net_value_from_db")
def get_net_value_from_db(data: GetNetValueModel, ph=Depends(pandahub)):
    return ph.get_net_value_from_db(**data.model_dump())

class SetNetValueModel(BaseCRUDModel):
    element_index: int
    parameter: str
    value: Any = None

@router.post("/set_net_value_in_db")
def set_net_value_in_db(data: SetNetValueModel, ph=Depends(pandahub)):
    return ph.set_net_value_in_db(**data.model_dump())

class CreateElementModel(BaseCRUDModel):
    element_index: int
    element_data: dict

@router.post("/create_element")
def create_element_in_db(data: CreateElementModel, ph=Depends(pandahub)):
    return ph.create_element(**data.model_dump())

class CreateElementsModel(BaseCRUDModel):
    elements_data: list[dict[str,Any]]

@router.post("/create_elements")
def create_elements_in_db(data: CreateElementsModel, ph=Depends(pandahub)):
    return ph.create_elements(**data.model_dump())

class DeleteElementModel(BaseCRUDModel):
    element_index: int

@router.post("/delete_element")
def delete_net_element(data: DeleteElementModel, ph=Depends(pandahub)):
    return ph.delete_element(**data.model_dump())

class DeleteElementsModel(BaseCRUDModel):
    element_indexes: list[int]

@router.post("/delete_elements")
def delete_net_elements(data: DeleteElementsModel, ph=Depends(pandahub)):
    return ph.delete_elements(**data.model_dump())

### deprecated routes
@router.post("/create_element_in_db")
def create_element_in_db(*args, **kwargs):
    raise RuntimeError("create_element_in_db was deprecated - use create_element instead!")

@router.post("/create_elements_in_db")
def create_elements_in_db(*args, **kwargs):
    raise RuntimeError("create_elements_in_db was deprecated - use create_elements instead!")


@router.post("/delete_net_element")
def delete_net_element(*args, **kwargs):
    raise RuntimeError("delete_net_element was deprecated - use delete_element instead!")



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 26 12:25:55 2023

@author: LeonThurner
"""

import pandapower as pp
from pandahub import PandaHub
import time
from line_profiler import LineProfiler

def get_test_net(n_buses):
    net = pp.create_empty_network()
    pp.create_buses(net, n_buses, vn_kv=0.4)
    pp.create_loads(net, net.bus.index[::50], p_mw=0, q_mvar=0)
    pp.create_sgens(net, net.bus.index[::50], p_mw=0, q_mvar=0)
    pp.create_gens(net, net.bus.index[::1000], p_mw=0, q_mvar=0)
    pp.create_lines(net, net.bus.index[::2], net.bus.index[1::2], 1.0, "NAYY 4x50 SE")
    pp.create_transformers_from_parameters(net, net.bus.index[::1000],
                                            net.bus.index[1::1000],
                                            vn_hv_kv=20.,
                                            vn_lv_kv=0.4,
                                            sn_mva=0.4,
                                            vkr_percent=5.,
                                            vk_percent=0.5,
                                            pfe_kw=0,
                                            i0_percent=0,
                                            )
    line_buses = list(net.line.from_bus.values) + list(net.line.to_bus.values)
    line_indices = list(net.line.index) + list(net.line.index)
    pp.create_switches(net, line_buses, line_indices, "l")
    return net

def write_test_net_to_mongodb(net, project_name):
    ph = PandaHub()
    if ph.project_exists(project_name):
        ph.set_active_project(project_name)
        ph.delete_project(True)
    ph.create_project(project_name)
    ph.write_network_to_db(net, "test_net")

def load_test_subnet_from_mongodb(project_name):
    ph = PandaHub()
    ph.set_active_project(project_name)
    buses = list(range(1000))
    t0 = time.time()
    subnet = ph.get_subnet_by_name("test_net",
                                   node_filter={"index": {"$in": buses}})
    t1 = time.time()
    print("LOADED NET IN %.2fs"%(t1-t0))
    return subnet

def profile_load_test(project_name):
    lp = LineProfiler()
    lp_wrapper = lp(load_test_subnet_from_mongodb)
    lp.add_function(PandaHub.get_subnet)
    lp.add_function(PandaHub._add_element_from_collection)
    lp_wrapper(project_name)
    lp.print_stats()

if __name__ == '__main__':
    n_buses = 3e6 #3 Million buses
    project_name = "test_%u"%n_buses
    net = get_test_net(n_buses)
    write_test_net_to_mongodb(net, project_name)
    profile_load_test(project_name)
    # subnet = load_test_subnet_from_mongodb(project_name)




DATATYPES = {
    "bus": {
        "name": str,
        "vn_kv": float,
        "type": str,
        "zone": str,
        "in_service": bool,
        "max_vm_pu": float,
        "min_vm_pu": float,
    },
    "load": {
        "name": str,
        "bus": int,
        "p_mw": float,
        "q_mvar": float,
        "const_z_percent": float,
        "const_i_percent": float,
        "sn_mva": float,
        "scaling": float,
        "in_service": bool,
        "controllable": bool,
        "type": str
    },
    "sgen": {
        "name": str,
        "bus": int,
        "p_mw": float,
        "q_mvar": float,
        "sn_mva": float,
        "scaling": float,
        "in_service": bool,
        "type": str,
        "current_source": bool
    },
    "gen": {
        "name": str,
        "bus": int,
        "p_mw": float,
        "vm_pu": float,
        "sn_mva": float,
        "min_q_mvar": float,
        "max_q_mvar": float,
        "scaling": float,
        "slack": bool,
        "in_service": bool,
        "type": str,
        "slack_weight": float,
        "controllable": bool,
        "max_p_mw": float,
        "min_p_mw": float
    },
    "switch": {
        "name": str,
        "bus": int,
        "element": int,
        "et": str,
        "type": str,
        "closed": bool,
        "z_ohm": float
    },
    "ext_grid": {
        "name": str,
        "bus": int,
        "vm_pu": float,
        "va_degree": float,
        "slack_weight": float,
        "in_service": bool,
        "max_p_mw": float,
        "min_p_mw": float,
        "max_q_mvar": float,
        "min_q_mvar": float,
        "s_sc_max_mva": float,
        "s_sc_min_mva": float,
        "rx_max": float,
        "rx_min": float,
        "x0x_max": float,
        "x0x_min": float,
        "r0x0_max": float,
        "r0x0_min": float,
        "controllable": bool
    },
    "line": {
        "name": str,
        "from_bus": int,
        "to_bus": int,
        "length_km": float,
        "r_ohm_per_km": float,
        "x_ohm_per_km": float,
        "c_nf_per_km": float,
        "g_us_per_km": float,
        "max_i_ka": float,
        "df": float,
        "parallel": int,
        "type": str,
        "in_service": bool,
        "max_loading_percent": float,
        "std_type": str,
    },
    "trafo": {
        "name": str,
        "std_type": str,
        "hv_bus": int,
        "lv_bus": int,
        "sn_mva": float,
        "vn_hv_kv": float,
        "vn_lv_kv": float,
        "vk_percent": float,
        "vkr_percent": float,
        "pfe_kw": float,
        "i0_percent": float,
        "shift_degree": float,
        "tap_side": str,
        "tap_neutral": int,
        "tap_min": int,
        "tap_max": int,
        "tap_step_percent": float,
        "tap_step_degree": float,
        "tap_pos": int,
        "tap_phase_shifter": bool,
        "parallel": int,
        "df": float,
        "in_service": bool
    },
    "motor": {
        "name": str,
        "bus": int,
        "pn_mech_mw": float,
        "loading_percent": float,
        "cos_phi": float,
        "cos_phi_n": float,
        "efficiency_percent": float,
        "efficiency_n_percent": float,
        "lrc_pu": float,
        "vn_kv": float,
        "scaling": float,
        "in_service": bool,
        "rx": float
    },
    "asymmetric_load": {
        "name": str,
        "bus": int,
        "p_a_mw": float,
        "q_a_mvar": float,
        "p_b_mw": float,
        "q_b_mvar": float,
        "p_c_mw": float,
        "q_c_mvar": float,
        "sn_mva": float,
        "scaling": float,
        "in_service": bool,
        "type": str
    },
    "asymmetric_sgen": {
        "name": str,
        "bus": int,
        "p_a_mw": float,
        "q_a_mvar": float,
        "p_b_mw": float,
        "q_b_mvar": float,
        "p_c_mw": float,
        "q_c_mvar": float,
        "sn_mva": float,
        "scaling": float,
        "in_service": bool,
        "type": str,
        "current_source": bool
    },
    "storage": {
        "name": str,
        "bus": int,
        "p_mw": float,
        "q_mvar": float,
        "sn_mva": float,
        "soc_percent": float,
        "min_e_mwh": float,
        "max_e_mwh": float,
        "scaling": float,
        "in_service": bool,
        "type": str
    },
    "shunt": {
        "bus": int,
        "name": str,
        "q_mvar": float,
        "p_mw": float,
        "vn_kv": float,
        "step": int,
        "max_step": int,
        "in_service": bool
    },
    "trafo3w": {
        "name": str,
        "std_type": str,
        "hv_bus": int,
        "mv_bus": int,
        "lv_bus": int,
        "sn_hv_mva": float,
        "sn_mv_mva": float,
        "sn_lv_mva": float,
        "vn_hv_kv": float,
        "vn_mv_kv": float,
        "vn_lv_kv": float,
        "vk_hv_percent": float,
        "vk_mv_percent": float,
        "vk_lv_percent": float,
        "vkr_hv_percent": float,
        "vkr_mv_percent": float,
        "vkr_lv_percent": float,
        "pfe_kw": float,
        "i0_percent": float,
        "shift_mv_degree": float,
        "shift_lv_degree": float,
        "tap_side": str,
        "tap_neutral": int,
        "tap_min": int,
        "tap_max": int,
        "tap_step_percent": float,
        "tap_step_degree": float,
        "tap_pos": int,
        "tap_at_star_point": bool,
        "in_service": bool,
    },
    "impedance": {
        "name": str,
        "from_bus": int,
        "to_bus": int,
        "rft_pu": float,
        "xft_pu": float,
        "rtf_pu": float,
        "xtf_pu": float,
        "sn_mva": float,
        "in_service": bool
    },
    "dcline": {
        "name": str,
        "from_bus": int,
        "to_bus": int,
        "p_mw": float,
        "loss_percent": float,
        "loss_mw": float,
        "vm_from_pu": float,
        "vm_to_pu": float,
        "max_p_mw": float,
        "min_q_from_mvar": float,
        "min_q_to_mvar": float,
        "max_q_from_mvar": float,
        "max_q_to_mvar": float,
        "in_service": bool
    },
    "ward": {
        "name": str,
        "bus": int,
        "ps_mw": float,
        "qs_mvar": float,
        "qz_mvar": float,
        "pz_mw": float,
        "in_service": bool
    },
    "xward": {
        "name": str,
        "bus": int,
        "ps_mw": float,
        "qs_mvar": float,
        "qz_mvar": float,
        "pz_mw": float,
        "r_ohm": float,
        "x_ohm": float,
        "vm_pu": float,
        "slack_weight": float,
        "in_service": bool
    },
    "measurement": {
        "name": str,
        "measurement_type": str,
        "element_type": str,
        "element": int,
        "value": float,
        "std_dev": float,
        "side": str
    },
    "pwl_cost": {
        "power_type": str,
        "element": int,
        "et": str,
        "points": str
    },
    "poly_cost": {
        "element": int,
        "et": str,
        "cp0_eur": float,
        "cp1_eur_per_mw": float,
        "cp2_eur_per_mw2": float,
        "cq0_eur": float,
        "cq1_eur_per_mvar": float,
        "cq2_eur_per_mvar2": float
    },
    'characteristic': {
        'object': object
    },
    'controller': {
        'object': object,
        'in_service': bool,
        'order': float,
        'level': str,
        'initial_run': bool,
        "recycle": str
    },
}



import pytest
from pandahub import PandaHub
# from pandahub.lib.settings import pandahub_settings


@pytest.fixture(scope="session")
def ph():
    ph = PandaHub()

    project_name = "pytest"

    if ph.project_exists(project_name):
        ph.set_active_project(project_name)
        ph.delete_project(i_know_this_action_is_final=True)

    ph.create_project(name=project_name)

    yield ph

    ph.delete_project(i_know_this_action_is_final=True)


